from typing import Callable

from pyray import Rectangle, Vector2, get_screen_height, get_screen_width # type: ignore


Map = Callable[[float], float]

class Slot:
	def __init__(self, x_map: Map, y_map: Map):
		self.x = x_map
		self.y = y_map

	def v(self, v: Vector2) -> Vector2:
		return Vector2(self.x(v.x), self.y(v.y))
	
	def __enter__(self): return self
	def __exit__(self, exc_type, exc_value, trace): pass

	def vxy(self, x: float, y: float) -> Vector2:
		return self.v(Vector2(x, y))
	
	def r(self, r: Rectangle=Rectangle(0,0,1,1)) -> Rectangle:
		x0 = self.x(r.x)
		y0 = self.y(r.y)
		x1 = self.x(r.x+r.width)
		y1 = self.y(r.y+r.height)
		x_min = min(x0, x1)
		y_min = min(y0, y1)
		x_max = max(x0, x1)
		y_max = max(y0, y1)
		return Rectangle(x_min, y_min, x_max-x_min, y_max-y_min)

	@staticmethod
	def get_root_slot(*, in_uv:bool=True) -> 'Slot':
		"""
		Returns the root slot for the window
		"""
		screen_width = get_screen_width()
		screen_height = get_screen_height()
		if in_uv:
			return Slot(lambda x: x*screen_width, lambda y: y*screen_height)
		else:
			return Slot(lambda x: x, lambda y: y)
	
	def between(self, x0: float, y0: float, x1: float, y1: float) -> 'Slot':
		return Slot(lambda x: self.x(x*(x1-x0)+x0), lambda y: self.y(y*(y1-y0)+y0))
	
	def remap(self, x0: float, y0: float, x1: float, y1: float) -> 'Slot':
		"""
		Returns the slot with the same region on the window but using a different coordinate space
		"""
		return Slot(lambda x: self.x((x-x0)/(x1-x0)), lambda y: self.y((y-y0)/(y1-y0)))
	
	def grid(self, rows: int, cols: int, *, total_slots:int=-1) -> 'list[Slot]':
		return self.extending_grid(rows, cols, rows, total_slots=total_slots)
	
	def extending_grid(self, rows: int, cols: int, total_rows: int, *, total_slots:int=-1) -> 'list[Slot]':
		slots: list[Slot] = []
		if total_rows == -1:
			assert total_slots != -1
			total_rows = (total_slots+cols-1)//cols
		for j in range(total_rows):
			for i in range(cols):
				slots.append(self.between(i/cols, j/rows, (i+1)/cols, (j+1)/rows))
		if total_slots != -1: return slots[:total_slots]
		return slots

	
	def translate(self, x0: float, y0: float) -> 'Slot':
		return self.between(x0, y0, x0+1, y0+1)

	
	def square(self) -> 'Slot':
		"""
		Returns the square slot that takes up as much space as possible within parent slot
		"""
		return self.with_ratio(1)
		
	def with_ratio(self, width_to_height_ratio: float) -> 'Slot':
		width = self.x(1) - self.x(0)
		height = self.y(1) - self.y(0)

		if width == 0 or height == 0: return Slot(lambda x: self.x(0) + width/2, lambda y: self.y(0) + height/2)

		if width/height > width_to_height_ratio:
			ratio = height/width*width_to_height_ratio
			left = (1-ratio)/2
			return self.between(left, 0, 1-left, 1)
		else:
			ratio = width/height/width_to_height_ratio
			up = (1-ratio)/2
			return self.between(0, up, 1, 1-up)
		
	def padding(self, horz_padding: float, vert_padding: float, *, width:float=1, height:float=1) -> 'Slot':
		return self.between(horz_padding, vert_padding, width-horz_padding, height-horz_padding)

